/***************************************************************************
 *
 * Copyright (c) 1999 Balzs Scheidler
 * Copyright (c) 1999 BalaBit Computing
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Inspired by nsyslog, originally written by Darren Reed.
 *
 * $Id$
 *
 ***************************************************************************/

#include "afsocket.h"
#include "cfgfile.h"
#include "format.h"
#include "pkt_buffer.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "afunix.c.x"

/* af_unix source connection */

static int
do_init_afunix_connection(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afsocket_source_connection, self, c);
	struct close_callback *cb;
	struct resource_node *res;

	res = REMEMBER_RESOURCE(cfg->resources, &self->client->super.super);
	cb = make_afsocket_source_close_callback(res, cfg);

	io_read(self->client, make_log_reader(1, c), cb);
	
	return ST_OK | ST_GOON;
}

static void do_handle_afunix_line(struct log_handler *c, struct log_info *msg)
{
	CAST(afsocket_source_connection, self, c);

	if (msg->host) {
		ol_string_free(msg->host);
		msg->host = NULL;
	}
	msg->flags |= LF_LOCAL;

	HANDLE_LOG(self->super.next, msg);
}

struct log_handler *
make_afunix_source_connection(struct io_fd *client, 
			      struct address_info *client_addr,
			      struct log_handler *next)
{
	NEW(afsocket_source_connection, self);

	self->super.super.init = do_init_afunix_connection;
	self->super.super.handler = do_handle_afunix_line;
	self->super.next = next;
	self->client = client;
	self->client_addr = client_addr;
	return &self->super.super;
}

/* CLASS:
     (class
       (name afunix_source)
       (super afsocket_source)
       (vars
         (owner string)
	 (group string)
	 (perm simple int)
	 (cfg object syslog_conf)))
*/

void afunix_set_owner(struct log_source_driver *c, const char *owner)
{
	CAST(afunix_source, self, c);

	if (self->owner)
		ol_string_free(self->owner);
	self->owner = c_format_cstring("%z", owner);
}

void afunix_set_group(struct log_source_driver *c, const char *group)
{
	CAST(afunix_source, self, c);

	if (self->group)
		ol_string_free(self->group);
	self->group = c_format_cstring("%z", group);
}

void afunix_set_perm(struct log_source_driver *c, int perm)
{
	CAST(afunix_source, self, c);

	self->perm = perm;
}

static int
do_open_unix_connection(struct afsocket_source *c,
			struct io_fd *client,
			struct address_info *client_addr)
{
	CAST(afunix_source, self, c);
	struct log_handler *conn;

	conn = make_afunix_source_connection(client, client_addr, self->super.super.super.next);
	
	return LOG_HANDLER_INIT(conn, self->cfg);
}

int do_init_afunix_source(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afunix_source, self, c);
	CAST(unix_address_info, bind_addr, self->super.bind_addr);
	struct group *grp = NULL;
	struct passwd *pwd = NULL;
	int fd;

	self->cfg = cfg;
	fd = io_open_socket(bind_addr->super.family, 
			    self->super.flags & AFSOCKET_DGRAM ? SOCK_DGRAM : SOCK_STREAM,
			    0, 
			    &bind_addr->super);
	if (fd < 0) {
		return ST_QUIT | ST_FAIL;
	}

	if (self->owner)
		pwd = getpwnam(self->owner->data);
	if (self->group)
		grp = getgrnam(self->group->data);
	chown(bind_addr->path->data, pwd ? pwd->pw_uid : -1, grp ? grp->gr_gid : -1);
	chmod(bind_addr->path->data, self->perm);

	switch (self->super.flags & 0x0003) {
	case AFSOCKET_STREAM: {
		struct listen_fd *listen;
		if ((listen = 
		     io_listen(cfg->backend, fd,
			       make_afsocket_accept_callback
			       (cfg->backend, &self->super)))) {
			REMEMBER_RESOURCE(cfg->resources, &listen->super.super);
		}
		else {
			close(fd);
			return ST_FAIL | ST_QUIT;
		}
		break;
	}
			
	case AFSOCKET_DGRAM: {
		struct log_handler *conn;
		struct io_fd *client = make_io_fd(cfg->backend, fd);
					
		conn = make_afunix_source_connection(client, NULL, c);
		LOG_HANDLER_INIT(conn, cfg);

		break;
	}
	default:
		return ST_FAIL | ST_QUIT;
	}

	return ST_OK | ST_GOON;		
}

struct log_source_driver *
make_afunix_source(struct address_info *bind_addr, UINT32 flags)
{
        NEW(afunix_source, self);

        self->super.super.super.super.init = do_init_afunix_source;
	self->super.bind_addr = bind_addr;
        self->super.flags = flags;
	self->super.open_connection = do_open_unix_connection;
	self->perm = 0666;
	
        return &self->super.super;
}


/* CLASS:
   (class
     (name afunix_dest)
     (super afsocket_dest)
     (vars
       (dest object io_fd)
       (dest_buf object abstract_buffer)
       (cfg object syslog_conf)))
*/

void do_handle_afunix_dest(struct log_handler *c, struct log_info *msg)
{
	CAST(afunix_dest, self, c);

	if (self->dest_buf) {
		A_WRITE_STRING(&self->dest_buf->super, c_format("<%i>%S %S %S%c", msg->pri, msg->date, msg->host, msg->msg, 0));
	}
	log_info_free(msg);
}

static int do_init_afunix_dest(struct log_handler *c, struct syslog_conf *cfg)
{
	CAST(afunix_dest, self, c);
	struct connect_fd *conn_fd;
	int fd;

	self->cfg = cfg;
	if (self->super.super.log_fifo_size == -1)
		self->super.super.log_fifo_size = cfg->log_fifo_size;
	
	fd = io_open_socket(self->super.dest_addr->family, 
			    (self->super.flags & AFSOCKET_DGRAM) 
			    ? SOCK_DGRAM : SOCK_STREAM, 
			    0,
			    self->super.bind_addr);
	
	if (fd < 0) {
		return ST_FAIL | ST_QUIT;
	}
	if (!self->dest_buf)
		self->dest_buf = make_pkt_buffer(self->super.super.log_fifo_size);

	conn_fd = io_connect(cfg->backend, 
			     fd, 
			     self->super.dest_addr, 
			     make_afsocket_dest_connected(cfg->backend, &self->super));

	if (conn_fd) {
		return ST_OK | ST_GOON;
	}
	else {
		CAST(unix_address_info, unix_addr, self->super.dest_addr);

		close(fd);
		werror("Error connecting AF_UNIX socket (%S): %z\n", unix_addr->path, strerror(errno));
	}
	return ST_FAIL | ST_QUIT;
}

static int
do_afunix_dest_connected(struct afsocket_dest *c,
			 struct io_fd *server)
{
	CAST(afunix_dest, self, c);
	struct resource_node *n;

	if (self->dest)
		self->dest->buffer = NULL;
	if (server) {
		n = REMEMBER_RESOURCE(self->cfg->resources, 
				      &server->super.super);
		self->dest = io_write(server, 
				      self->dest_buf,
				      make_dest_close(&self->super.super.super, n, self->cfg->time_reopen, self->cfg));
	}
	else {
		CAST(unix_address_info, unix_addr, self->super.dest_addr);

		werror("Error connecting to remote AF_UNIX socket (%S), reattempting in %i seconds\n", unix_addr->path, self->cfg->time_reopen);
		self->dest = NULL;
		io_callout(self->cfg->backend, self->cfg->time_reopen, make_dest_reopen(&self->super.super.super, self->cfg));
	}
	return ST_OK | ST_GOON;

}

struct log_dest_driver *make_afunix_dest(struct address_info *dest_addr, UINT32 flags)
{
	NEW(afunix_dest, self);
	
	self->super.super.super.init = do_init_afunix_dest;
	self->super.super.super.handler = do_handle_afunix_dest;
	self->super.super.log_fifo_size = -1;
	self->super.connected = do_afunix_dest_connected;
	self->super.flags = flags;
	self->super.dest_addr = dest_addr;
	return &self->super.super;
}

